\ Read mail from POP3 server and store data
\ Part of RS-mail system
\ Ver. 2.1 (partial MIME support)
\ Pavel Zampach, 2022
\ *

inlib fileio				\ file I/O support
inlib statmem				\ static memory management
inlib strcon				\ string constant
inlib strvar				\ string variable
inlib buflib				\ buffer:
inlib applib				\ append string
inlib base64				\ base64 encoding/decoding
inlib qprint				\ quoted-printable encoding/decoding
inlib charlib				\ charset conversions
inlib modemrcv				\ modem receive routines
inlib timedate				\ TIME&DATE (for DBFACC)
inlib dbfconr				\ field constants for RMAIL.DBF
inlib dbfacc				\ dBase II access
inlib ccq				\ C,"
cr ." RMAIL 2.1 loading"

decimal

255 constant B-LEN
b-len 1+ buffer: L-BUF			\ basic line buffer
b-len 1+ buffer: X-BUF			\ help buffer

50000. 2constant MAXLEN			\ length of msg for full processing
s" B0:$$$.MSG" constr WRK-NAME		\ work file in fast RAM disk
s" G0:" constr ATT-PATH			\ path for attachment saving

\ Mail account specific data
s" USER name@domain.com" constr MUSER	\ user name for POP3 server
s" PASS password" constr MPASS		\ password for POP3 server
s" mail.domain.com:110" constr MURL	\ URL of POP3 server

variable ACTMSG				\ number of actual msg
variable FNBASE				\ number base for msg body filename
variable MULTI-FLAG			\ flag of multipart msg
variable WRITE-CR			\ flag of writing CRLF to output file
variable OUT-OPEN			\ flag of open output
variable #OUT-FILE			\ number of output files
integer FID-O				\ output file ID
128 string: BOUNDARY			\ boundary
max-path string: F-NAME			\ output filename
char ? constant "?"

\ --- general support words------------

\ Charset conversion
defer CHAR-CONV	( addr u1 -- addr u2 )

\ Transfer decoding
defer TRAN-CONV	( addr1 u1 addr2 u2 -- addr1 u3 )

\ Substring [addr2 u2] from string [addr1 u1] delimited by char [c1]
\ (begin) and [c2] (end). Delimiter chars are not included.
: /QUOT		( addr1 u1 c1 c2 -- addr2 u2 )
  >r scan 1 /string
  over swap r> scan drop over -
;

\ Return the first substring [addr3 u3] from string [addr1 u1]
\ separated by char [c], string [addr2 u2] is the remainder. When
\ [addr3 u3] is the last part of original string, u2 = -1
: SLICE		( addr1 u1 c  -- addr2 u2 addr3 u3 )
  >r 2dup r> scan dup >r
  cskip 2swap r> -
;

\ Compare is Caps independent
: (STRING-OF)	( addr1 u1 addr2 u2 -- addr1 u1 flag )
  2over caps compare 0=
;

system
: STRING-OF	( addr1 u1 addr2 u2 -- ) ( C: -- orig )
  postpone (string-of)
  postpone if
  postpone 2drop
; immediate
application

\ [flag] is TRUE if char [c] is digit
: DIGIT?	( c -- flag )
  [char] 0 [char] 9 between
;

\ Scan string [addr1 u], [addr2] is address of the first digit (zero if no
\ digit was found)
: SCAN-DIG	( addr1 u -- addr2 )
  bounds do
    i c@ digit? if i unloop end
  loop   0
;

\ Decode "Encoded-word" string [addr1 u1] to string [addr2 u2] (in PAD)
: DEC-WORD	( addr1 u1 -- addr2 u2 )
  ['] noop is char-conv
  "?" slice 2drop
  "?" slice 2dup upper
  cond
    iso1 string-of ['] iso1>ascii is char-conv endof
    iso2 string-of ['] iso2>ascii is char-conv endof
    utf  string-of ['] utf>ascii  is char-conv endof
    win  string-of ['] win>ascii  is char-conv endof
    2drop
  cont
  "?" slice drop c@ upcase
  [char] B = if
    ['] app-base64d
  else
    undersc on
    ['] app-qprintd
  then
  >r
  "?" slice 2nip
  pad 0 2swap r> execute
  char-conv
  undersc off
;

\ L-BUF as string [addr u]
: L-STR		( -- addr u )
  l-buf count
;

\ X-BUF as string [addr u]
: X-STR		( -- addr u )
  x-buf count
;

\ ---support for communication------------

\ End of session with [addr u] msg
: END-SES	( addr u -- )
  type cr
  flushport
  s" QUIT" send-line flushport		\ QUIT and read the rest of data
  mreset
  quit
;

\ Extended timeout check
:NONAME		( ior -- )
  if s" time out" end-ses then
; is tc

\ [flag] is TRUE if the first char of response is "+"
: RESP-OK	( -- flag )
  l-buf rcv-line
  l-buf 1+ c@ [char] + =
;

\ Return string [addr u] with data from server response
: RESP-DATA	( -- addr u )
  l-str 4 /string			\ skip over "+OK "
;

\ Read line (end by LF) to L-BUF and test response
: LINE-TEST	( -- )
  resp-ok not if			\ -ERR|+OK
    l-str end-ses
  then
;

\ Send command [addr u] supplemented by actual msg #
: SEND-ACT	( addr u -- )
  >pad appbl actmsg @ appn send-line
;

\ ---communications with POP3 server----

\ Connect to POP3 server
: CONNECT	( -- )
  mreset mquiet
  500 to wticks
  murl mdial
  begin resp-ok until
  20 to wticks
  muser send-line line-test		\ +OK
  mpass send-line line-test		\ +OK
  ." Connected" cr
;

\ Return number [n] of messages
: #MSG		( -- n )
  s" STAT" send-line			\ STAT
  line-test resp-data			\ +OK data
  /number 2nip d>s			\ fetch number from response
;

\ Return actual message length [d]
: MSG-LEN	( -- d )
  s" LIST" send-act			\ LIST am
  line-test resp-data			\ +OK data
  bl scan 1 /string /number 2nip	\ fetch number from response
;

\ Delete actual message
: MSG-DEL	( -- )
  s" DELE" send-act			\ DELE am
  line-test				\ +OK
;

\ RETR actual message
: MSG-RETR	( -- )
  s" RETR" send-act			\ RETR am
  line-test				\ +OK
;

\ TOP actual message for [n] lines
: MSG-TOP	( n -- )
  >r
  s" TOP " >pad
  actmsg @ appn appbl r> appn send-line	\ TOP am n
  line-test				\ +OK
;

\ Read whole msg from modem and save it to work file
: MSG-READ	( -- )
  wrk-name open-data
  msg-len maxlen d< if
    msg-retr
  else
    50 msg-top
  then
  rcv-data
  close-data
;

\ ---support for parsing------------------

\ Search (case insensitive) string [addr u] in L-BUF
: L-SEARCH	( addr u -- flag )
  l-str 2swap caps search nip nip
;

\ Read line from input file to L-BUF as counted string
\ [flag] is TRUE if read data is valid
: L-LINE?	( -- flag )
  l-buf 1+ b-len fid read-line fc
  swap l-buf c!
;

\ Append string [addr u] to L-BUF as counted string
: +L-STR	( addr u -- )
  l-str +string swap 1- c!
;

\ Quoted part [addr u] of L-STR
: L-QUOT	( -- addr u )
  l-str [char] " dup /quot
;

\ Read and test for the last line in msg
\ [flag] is TRUE if read data is valid and it isn't the last line
: NOT-LEAST?	( -- flag )
  l-line?
  l-buf @ $2e01 <> and			\ line is not "."
;

\ [flag] is TRUE if string [addr u] is next line of folded field
: FOLD?		( addr u -- flag)
  drop c@ [char] ! <
;

\ [c] = out-file# >=2 -> 'A', 1 -> '@'
: FLAG-CHAR	( -- c )
  #out-file @ 2 min [char] ? +
;

\ Header field name [addr u] from L-STR
: FIELD-NAME	( -- addr u )
  l-str over swap [char] : scan drop over -
;

\ Header field body [addr u] from L-STR
: FIELD-BODY	( -- addr u )
  l-str -path -blanks
;

\ Filename as "MI####.MSG" [addr u] of msg body file
: MSG-NAME	( -- addr u )
  fnbase @ actmsg @ + s>d
  <# s" .MSG" shold # # # # s" MI" shold #>
;

\ Init FNBASE variable
: FNBASE-INIT	( -- )
  #recs dup if
    1- read-rec				\ read the last record
    FILE 2 /string /number 2nip d>s	\ extract number
  then
  fnbase !
;

\ Filename as "X0:MI####.AT#" [addr u] of msg attachment (in PAD)
: ATT-NAME	( -- addr u )
  att-path >pad
  msg-name 3 - apps
  s" AT" apps
  #out-file @ [char] @ + appc
;

create TAB-MO c," JanFebMarAprMayJunJulAugSepOctNovDec"

\ Convert 3-char month starting on [addr] to number 1..12 [n], returns
\ zero, if searching failed
: CONV-MO	( addr -- n )
  3 tab-mo 36 2swap caps search if
    drop tab-mo - 3 / 1+
  else
    2drop 0
  then
;

\ Convert [d] to two digit string
: TWO-DIG	( d -- addr u )
  <# # # #>
;

\ Decode string [addr1 u1] containing "Encode words" to string [addr2 u2]
\ (in L-BUF)
: DEC-STRING	( addr1 u1 -- addr2 u2 )
  over >r				\ save start addr
  l-buf off				\ init L-BUF
  begin
    s" =?" search
  while
    swap r> 2dup - +l-str swap		\ append string before "=?"
    2dup dec-word +l-str		\ append decoded "encoded word"
    s" ?=" search not if
      2drop l-str
    end
    2 /string   over >r			\ skip "?=" and save addr
  repeat   rdrop
  l-str +string				\ append rest
;

\ Read and append next lines if begin with LWSP to L-STR
: UNFOLD	( -- )
  begin
    fid file-position fc		\ mark file position
    x-buf 1+ b-len fid read-line fc	\ read line to X-BUF
    over x-buf c!			\ as counted string
    and	x-str fold? and			\ not empty line and test folding
  while
    2drop				\ drop file position
    x-str >blanks x-str bl skip		\ trim leading spaces and tabs
    -1 /string +l-str			\ one space back and append to L-STR
  repeat
  fid reposition-file fc		\ restore file position
;

\ ---parsing of work file--------------

\           1         2         3
\ 0123456789012345678901234567890123456789
\ Date: Tue, 14 Dec 2021 14:24:53 +0100
\ Date: Wed, 3 Nov 2021 11:46:01 +0100

\ Process "Date" field
: PRO-DATE	( -- )
  DATE [char] / fill			\ fill DATE field by "/"
  l-str scan-dig 2 /number two-dig	\ build day
  [ DATE get-fielda ] literal
  swap cmove				\ day to string
  drop 1+ dup conv-mo			\ build month
  s>d two-dig
  [ DATE get-fielda 3 + ] literal
  swap cmove				\ month to string
  6 +
  [ DATE get-fielda 6 + ] literal
  2 cmove				\ year to string
;

\ Process "Subject" field
: PRO-SUBJ	( -- )
  unfold
  field-body x-buf pack
  count dec-string
  SUBJECT set-fieldcl
;

\ Process "From" field
: PRO-FROM	( -- )
  unfold
  field-body [char] < [char] > /quot
  MAILADD set-fieldcl
;

\ Process "Content-Type" field
: PRO-CTYPE	( -- )
  unfold
  s" TEXT/PLAIN" l-search if
    out-open on
    msg-name f-name s!
  then
  s" BOUNDARY=" l-search if
    l-quot boundary s!
    multi-flag on
  then
  s" APPLICATION" l-search if
    out-open on
    att-name f-name s!
  then
  iso1 l-search if ['] iso1>ascii is char-conv end
  iso2 l-search if ['] iso2>ascii is char-conv end
  utf  l-search if ['] utf>ascii  is char-conv end
  win  l-search if ['] win>ascii  is char-conv end
;

\ Process "Content-Transfer-Encoding" field
: PRO-CTRANS	( -- )
  field-body
  s" QUOTED-PRINTABLE" string-of
    ['] app-qprintd is tran-conv
  end
  s" BASE64" string-of
    ['] app-base64d is tran-conv
    write-cr off
  end
  2drop
;

\ Process "Content-Disposition" field
: PRO-CDISP	( -- )
  unfold
  s" ATTACHMENT" l-search if
    out-open on
    att-name f-name s!
    [char] * emit
  then
  s" FILENAME" l-search if
    l-quot dup 13 < if			\ len < 13 -> DOS (8.3) name
      att-path >pad +string
      2dup upper f-name s!
    else
      2drop
    then
  then
;

\ Process header line in L-STR
: HEAD-LINE	( -- )
  l-str fold? if end			\ skip next part of folded fields
  field-name
  s" DATE" string-of pro-date end
  s" SUBJECT" string-of pro-subj end
  s" FROM" string-of pro-from end
  s" CONTENT-TYPE" string-of pro-ctype end
  s" CONTENT-TRANSFER-ENCODING" string-of pro-ctrans end
  s" CONTENT-DISPOSITION" string-of pro-cdisp end
  2drop
;

\ Process primary or secondary header
: PRO-HDR	( -- )
  out-open off
  write-cr on
  ['] apps is tran-conv			\ no decoding
  ['] noop is char-conv			\ no char conversion
  begin
    l-line? l-buf c@ and		\ not-empty line
  while
    head-line
  repeat
  out-open @ if
    f-name s@ w/o create-file fc to fid-o
    1 #out-file +!
  then
;

\ Close output file
: OUT-CLOSE	( -- )
  out-open @ if
    fid-o close-file fc
    out-open off
  then
;

\ ---main processing routines----------

\ Process body line
: BODY-LINE	( -- )
  out-open @ if
    x-buf 0 l-str tran-conv		\ transfer decoding
    char-conv				\ charset conversion
    fid-o
    write-cr @ if write-line else write-file then
    fc
  then
;

\ Process ONEPART msg body
: ONE-MSG	( -- )
  out-open @ not if			\ if non-MIME old plain msg
    msg-name w/o create-file fc to fid-o
    out-open on
  then
  begin
    not-least?
  while
    body-line
  repeat
  out-close
;

\ Process MULTIPART msg body
: MULTI-MSG	( -- )
  begin
    not-least?
  while
    boundary s@ l-search if		\ boundary?
      out-close				\ close file if opened in previous part
      pro-hdr				\ process secondary header
    else
      body-line
    then
  repeat
  out-close
;

\ ---main part-------------------------

\ Process actual message
: PRO-MSG	( -- )
  ." Loading mail #" actmsg ?
  msg-read				\ read whole msg to work file
  msg-del
  #out-file off
  multi-flag off
  wrk-name r/o open-file fc to fid
  pro-hdr				\ process header
  multi-flag @ if
    multi-msg
  else
    one-msg
  then
  fid close-file fc
  wrk-name delete-file fc
  flag-char FLAG get-fielda c!
  msg-name FILE set-fieldcl
  append-rec
  ." from: " MAILADD get-fieldct type cr
;

: MAIN		( -- )
  connect
  s" RMAIL" open-dbf
  fnbase-init
  #msg 0 ?do
    i 1+ actmsg !
    pro-msg
  loop
  close-dbf
  s" Quit" end-ses
;

turnkey main e0:rmail
